//
// Persistence-of-Vision Watch Prototype
//
// This file contains code for the Microchip PIC16LF628
//
// Last update: September 22, 2004
//
// Copyright (c) 2004 Grand Idea Studio, Inc. (http://www.grandideastudio.com)
//

#include <pic.h>
#include "stdio.h"
#include "string.h"

// Configuration Bits:
// Oscillator = INT RC - I/O ON RA6/RA7
// Watchdog Timer (WDT) = DISABLED
// Power Up Timer (PWRT) = ENABLED
// Brown-Out Detect (BOR) = DISABLED
// Master Clear = ENABLED
// Low Voltage Program = DISABLED
// Data EE Read Protect = DISABLED
// Code Protect = DISABLED
__CONFIG(INTIO & WDTDIS & PWRTEN & BORDIS & MCLREN & LVPDIS & UNPROTECT);  

const char copyright[] = "Copyright (c) 2004 Grand Idea Studio, Inc. [www.grandideastudio.com]";

// DEFINES

//#define REVERSE     1       // Define to reverse direction of display

#define INPUT       1		// Port directions
#define OUTPUT      0
#define ON          1		// LED control
#define OFF         0
#define HIGH        1		// Pin level
#define LOW         0

#define PORTBIT(adr, bit)   ((unsigned)(&adr)*8+(bit))

#define	delay_us(x)	{ unsigned char _dcnt; _dcnt = (x)/(44000/4000)|1; while(--_dcnt != 0) continue; }

// GLOBAL VARIABLES

static bit D1 @ PORTBIT(PORTA, 2);     // Output
static bit D2 @ PORTBIT(PORTA, 3);     // Output
static bit D3 @ PORTBIT(PORTA, 1);     // Output
static bit D4 @ PORTBIT(PORTB, 0);     // Output
static bit D5 @ PORTBIT(PORTB, 1);     // Output
static bit D6 @ PORTBIT(PORTB, 2);     // Output
static bit D7 @ PORTBIT(PORTB, 4);     // Output

static bit TILT @ PORTBIT(PORTA, 4);   // Input


/*--------------------------------------------------------------------------------------------------
                                     Character generator

             This table defines the standard ASCII characters in a 5x7 dot format.
--------------------------------------------------------------------------------------------------*/
const char FontLookup [][5] =
{
    { 0x00, 0x00, 0x00, 0x00, 0x00 },   // sp
    { 0x00, 0x00, 0x2f, 0x00, 0x00 },   // !
    { 0x00, 0x07, 0x00, 0x07, 0x00 },   // "
    { 0x14, 0x7f, 0x14, 0x7f, 0x14 },   // #
    { 0x24, 0x2a, 0x7f, 0x2a, 0x12 },   // $
    { 0xc4, 0xc8, 0x10, 0x26, 0x46 },   // %
    { 0x36, 0x49, 0x55, 0x22, 0x50 },   // &
    { 0x00, 0x05, 0x03, 0x00, 0x00 },   // '
    { 0x00, 0x1c, 0x22, 0x41, 0x00 },   // (
    { 0x00, 0x41, 0x22, 0x1c, 0x00 },   // )
    { 0x14, 0x08, 0x3E, 0x08, 0x14 },   // *
    { 0x08, 0x08, 0x3E, 0x08, 0x08 },   // +
    { 0x00, 0x00, 0x50, 0x30, 0x00 },   // ,
    { 0x10, 0x10, 0x10, 0x10, 0x10 },   // -
    { 0x00, 0x60, 0x60, 0x00, 0x00 },   // .
    { 0x20, 0x10, 0x08, 0x04, 0x02 },   // /
    { 0x3E, 0x51, 0x49, 0x45, 0x3E },   // 0
    { 0x00, 0x42, 0x7F, 0x40, 0x00 },   // 1
    { 0x42, 0x61, 0x51, 0x49, 0x46 },   // 2
    { 0x21, 0x41, 0x45, 0x4B, 0x31 },   // 3
    { 0x18, 0x14, 0x12, 0x7F, 0x10 },   // 4
    { 0x27, 0x45, 0x45, 0x45, 0x39 },   // 5
    { 0x3C, 0x4A, 0x49, 0x49, 0x30 },   // 6
    { 0x01, 0x71, 0x09, 0x05, 0x03 },   // 7
    { 0x36, 0x49, 0x49, 0x49, 0x36 },   // 8
    { 0x06, 0x49, 0x49, 0x29, 0x1E },   // 9
    { 0x00, 0x36, 0x36, 0x00, 0x00 },   // :
    { 0x00, 0x56, 0x36, 0x00, 0x00 },   // ;
    { 0x08, 0x14, 0x22, 0x41, 0x00 },   // <
    { 0x14, 0x14, 0x14, 0x14, 0x14 },   // =
    { 0x00, 0x41, 0x22, 0x14, 0x08 },   // >
    { 0x02, 0x01, 0x51, 0x09, 0x06 },   // ?
    { 0x32, 0x49, 0x59, 0x51, 0x3E },   // @
    { 0x7E, 0x11, 0x11, 0x11, 0x7E },   // A
    { 0x7F, 0x49, 0x49, 0x49, 0x36 },   // B
    { 0x3E, 0x41, 0x41, 0x41, 0x22 },   // C
    { 0x7F, 0x41, 0x41, 0x22, 0x1C },   // D
    { 0x7F, 0x49, 0x49, 0x49, 0x41 },   // E
    { 0x7F, 0x09, 0x09, 0x09, 0x01 },   // F
    { 0x3E, 0x41, 0x49, 0x49, 0x7A },   // G
    { 0x7F, 0x08, 0x08, 0x08, 0x7F },   // H
    { 0x00, 0x41, 0x7F, 0x41, 0x00 },   // I
    { 0x20, 0x40, 0x41, 0x3F, 0x01 },   // J
    { 0x7F, 0x08, 0x14, 0x22, 0x41 },   // K
    { 0x7F, 0x40, 0x40, 0x40, 0x40 },   // L
    { 0x7F, 0x02, 0x0C, 0x02, 0x7F },   // M
    { 0x7F, 0x04, 0x08, 0x10, 0x7F },   // N
    { 0x3E, 0x41, 0x41, 0x41, 0x3E },   // O
    { 0x7F, 0x09, 0x09, 0x09, 0x06 },   // P
    { 0x3E, 0x41, 0x51, 0x21, 0x5E },   // Q
    { 0x7F, 0x09, 0x19, 0x29, 0x46 },   // R
    { 0x46, 0x49, 0x49, 0x49, 0x31 },   // S
    { 0x01, 0x01, 0x7F, 0x01, 0x01 },   // T
    { 0x3F, 0x40, 0x40, 0x40, 0x3F },   // U
    { 0x1F, 0x20, 0x40, 0x20, 0x1F },   // V
    { 0x3F, 0x40, 0x38, 0x40, 0x3F },   // W
    { 0x63, 0x14, 0x08, 0x14, 0x63 },   // X
    { 0x07, 0x08, 0x70, 0x08, 0x07 },   // Y
    { 0x61, 0x51, 0x49, 0x45, 0x43 },   // Z
    { 0x00, 0x7F, 0x41, 0x41, 0x00 },   // [
    { 0x55, 0x2A, 0x55, 0x2A, 0x55 },   // 55
    { 0x00, 0x41, 0x41, 0x7F, 0x00 },   // ]
    { 0x04, 0x02, 0x01, 0x02, 0x04 },   // ^
    { 0x40, 0x40, 0x40, 0x40, 0x40 },   // _
    { 0x00, 0x01, 0x02, 0x04, 0x00 },   // '
    { 0x20, 0x54, 0x54, 0x54, 0x78 },   // a
    { 0x7F, 0x48, 0x44, 0x44, 0x38 },   // b
    { 0x38, 0x44, 0x44, 0x44, 0x20 },   // c
    { 0x38, 0x44, 0x44, 0x48, 0x7F },   // d
    { 0x38, 0x54, 0x54, 0x54, 0x18 },   // e
    { 0x08, 0x7E, 0x09, 0x01, 0x02 },   // f
    { 0x0C, 0x52, 0x52, 0x52, 0x3E },   // g
    { 0x7F, 0x08, 0x04, 0x04, 0x78 },   // h
    { 0x00, 0x44, 0x7D, 0x40, 0x00 },   // i
    { 0x20, 0x40, 0x44, 0x3D, 0x00 },   // j
    { 0x7F, 0x10, 0x28, 0x44, 0x00 },   // k
    { 0x00, 0x41, 0x7F, 0x40, 0x00 },   // l
    { 0x7C, 0x04, 0x18, 0x04, 0x78 },   // m
    { 0x7C, 0x08, 0x04, 0x04, 0x78 },   // n
    { 0x38, 0x44, 0x44, 0x44, 0x38 },   // o
    { 0x7C, 0x14, 0x14, 0x14, 0x08 },   // p
    { 0x08, 0x14, 0x14, 0x18, 0x7C },   // q
    { 0x7C, 0x08, 0x04, 0x04, 0x08 },   // r
    { 0x48, 0x54, 0x54, 0x54, 0x20 },   // s
    { 0x04, 0x3F, 0x44, 0x40, 0x20 },   // t
    { 0x3C, 0x40, 0x40, 0x20, 0x7C },   // u
    { 0x1C, 0x20, 0x40, 0x20, 0x1C },   // v
    { 0x3C, 0x40, 0x30, 0x40, 0x3C },   // w
    { 0x44, 0x28, 0x10, 0x28, 0x44 },   // x
    { 0x0C, 0x50, 0x50, 0x50, 0x3C },   // y
    { 0x44, 0x64, 0x54, 0x4C, 0x44 }    // z
};

static unsigned int hour, minute, second;
static unsigned char cnt = 0;

// FUNCTION PROTOTYPES

void SetClock (char *str);
void LcdChr (char ch);
void LcdStr (char *dataPtr);
void hardware_init (void);
void StrReverse(char* str);
void delay_ms (unsigned long cnt);

void main (void)
{
    //char str[] = "Keely     ";
    char str[16];

	hardware_init();
    hour = 0;      // initialize clock
    minute = 0;
    second = 0; 

    // system initialized and ready to go
    while (1)
    {
        hour = 0; // don't display hour for proof-of-concept
        SetClock (str); // update clock

#ifdef REVERSE
        StrReverse(str);   // reverse the string to account for direction of arm swing
#endif

#ifdef REVERSE
        if (TILT) LcdStr(str); // only display message in one direction to avoid writing in reverse
#else
        if (!TILT) LcdStr(str);
#endif

    }
}

void interrupt isr (void)
{
    if (TMR1IF)   // Timer 1 interrupt
    { 	
        if (cnt == 15) // TMR1 overflows every 0.065536 seconds (close enough for proof-of-concept)
        {
            if (second == 59)
            {
                second = 0;
                if (minute == 59)
                {
                    minute = 0;
                    if (hour == 99)
                    {
                        hour = minute = second = 0; // reset clock
                    }
                    else ++hour;
                }
                else ++minute;
            }
            else ++second;

            cnt = 0;
        }
        else
            ++cnt;

        TMR1IF = 0;     // clear the interrupt
    }
}

void SetClock (char *str)
{
    char minstr[3], secstr[3];

    // no "0" padding with PICC sprintf, so have to do it manually
    if (minute < 10)
        sprintf(minstr, "0%d", minute);
    else
        sprintf(minstr, "%d", minute);
    minstr[2] = '\0';

    if (second < 10)
        sprintf(secstr, "0%d", second);
    else
        sprintf(secstr, "%d", second);
    secstr[2] = '\0';

    if (hour == 0)
        sprintf(str, "%s:%s     ", minstr, secstr);
    else
        sprintf(str, "%d:%s:%s     ", hour, minstr, secstr);
}

/*--------------------------------------------------------------------------------------------------

  Name         :  LcdChr

  Description  :  Displays a character.

  Argument(s)  :  ch   -> Character to write.

  Return value :  None.

--------------------------------------------------------------------------------------------------*/
void LcdChr (char ch)
{
    int i;
    char byte;

    if ( (ch < 0x20) || (ch > 0x7a) ) // If non-printable character,
    {
        //  Convert to a blank.
        ch = 32;
    }

#ifdef REVERSE
    for ( i = 4; i >= 0; i-- )
#else
    for ( i = 0; i < 5; i++ )
#endif

    {
        byte = FontLookup[ch - 32][i];
        
        D7 = (bit)(byte & 0x01);
        D6 = (bit)(byte >> 1) & 0x01;
        D5 = (bit)(byte >> 2) & 0x01;
        D4 = (bit)(byte >> 3) & 0x01;
        D3 = (bit)(byte >> 4) & 0x01;
        D2 = (bit)(byte >> 5) & 0x01;
        D1 = (bit)(byte >> 6) & 0x01;

        delay_ms(1);
        delay_us(250);
    }

	D1 = OFF;
    D2 = OFF;
    D3 = OFF;
    D4 = OFF;
    D5 = OFF;
    D6 = OFF;
    D7 = OFF;

    delay_us(250); // insert a small space between characters
    delay_us(250);
    delay_us(250);
    delay_us(250);
}

/*--------------------------------------------------------------------------------------------------

  Name         :  LcdStr

  Description  :  Displays a string.

  Argument(s)  :  dataPtr -> Pointer to null terminated ASCII string to display.

  Return value :  None.

--------------------------------------------------------------------------------------------------*/
void LcdStr (char *dataPtr)
{
    while ( *dataPtr )
    {
        LcdChr( *dataPtr++ );
    }
}

void hardware_init (void) // Configure hardware to a known state
{
    // Oscillator
    OSCF = 1;       // INTOSC oscillator frequency (4MHz)
    PSA = 0;        // Prescaler assigned to Timer0
    PS2 = 1;        // Prescaler 1:256
    PS1 = 1;
    PS0 = 1;

    // Interrupts
    GIE = 1;        // enable global interrupts
    PEIE = 1;       // enable peripheral interrupts
    TMR1IE = 1;     // enable timer1 interrupt

    // Peripherals
    T0CS = 0;       // TMR0 clock source = internal CLKOUT
    TMR1CS = 0;     // TMR1 clock source = internal clock (Fosc/4)
    TMR1ON = 0;
    T1CKPS1 = 0;    // TMR1 Prescaler 1:1
    T1CKPS0 = 0;    
    T1OSCEN = 0;    // TMR1 Oscillator disable
    TMR2ON = 0;
    VREN = 0;       // disable VREF
    VROE = 0;
    SPEN = 0;       // disable serial port
    CMCON = 0x07;   // disbale analog comparator, all pins digital
  
    // Ports
    PORTA = 0x00;
    TRISA0 = INPUT;
    TRISA1 = OUTPUT;
    TRISA2 = OUTPUT;
    TRISA3 = OUTPUT;
    TRISA4 = INPUT;     // open-drain output
    TRISA5 = INPUT;     // MCLR/VPP, always input
    TRISA6 = OUTPUT;    // Depends on oscillator mode
    TRISA7 = INPUT;
    
    PORTB = 0x00;
    RBPU = 1;           // disable PORTB pull-ups
    TRISB0 = OUTPUT;
    TRISB1 = OUTPUT;
    TRISB2 = OUTPUT;
    TRISB3 = OUTPUT;
    TRISB4 = OUTPUT;
    TRISB5 = INPUT;     
    TRISB6 = INPUT;
    TRISB7 = INPUT;

    // LEDs
	D1 = OFF;
    D2 = OFF;
    D3 = OFF;
    D4 = OFF;
    D5 = OFF;
    D6 = OFF;
    D7 = OFF;

    TMR1ON = 1; // Enable counter
}

void StrReverse(char* str)
{
  int c, i, j;
   
  for (i=0, j=strlen(str)-1; i < j;i++, j--)
  {
    c = str[i];
    str[i] = str[j];
    str[j] = c;
  }
}

// Delay routine configured for 4MHz INTOSC clock
void delay_ms (unsigned long cnt)
{
	unsigned long	i;
	
	do 
	{
		i = 4;
		do 
		{
			delay_us (250);
		} while(--i);
	} while(--cnt);
}
